# -*- coding: utf-8 -*-
"""# Matplotlib-Based Plotting Routines

This submodule contains methods related to visualize some of the standard Markov state plots.

"""
__all__ = [
    'plot_ck_test',
    'plot_wtd',
]

from ._ck_test import plot_ck_test
from ._wtd import plot_wtd
